import { SortDescriptor } from '../sort-descriptor';
import { Preprocessors } from '../common.interfaces';
/**
 * A type that represents a function which compares two values and returns `1`, `0`, or `-1` (minus one).
 *
 * For more information, refer to the [`composeSortDescriptors`]({% slug api_kendo-data-query_composesortdescriptors %}) configuration.
 *
 * @example
 * ```ts-no-run
 * const isGreaterThan: Comparer = (a, b) => a > b ? 1 : -1;
 * isGreaterThan(42, 42); // 0
 * isGreaterThan(22, 42); // -1
 * isGreaterThan(42, 22); // 1
 * ```
 */
export type Comparer = <T>(a: T, b: T) => number;
/**
 * Converts the `SortDescriptors` into a [`Comparer`]({% slug api_kendo-data-query_comparer %}) function that can be used through `Array.sort`. If multiple descriptors are provided, sorting is applied in a right-to-left order.
 * @param {SortDescriptor[]} descriptors - The descriptors which will be converted.
 * @returns {Comparer} - The produced function.
 *
 * @example
 * ```ts-no-run
 * import { composeSortDescriptors } from '@progress/kendo-data-query';
 *
 * const data = [{ name: "Pork" }, { name: "Pepper" }, { name: "Beef" } ];
 * const comparer = composeSortDescriptors([{ field: "name", dir: "asc" }]);
 * const result = data.sort(comparer);
 * // output: [{ name: "Beef" }, { name: "Pepper" }, { name: "Pork" }];
 * ```
 */
export declare const composeSortDescriptors: <T>(descriptors: SortDescriptor[], preprocessors?: Preprocessors<T>) => Comparer;
//# sourceMappingURL=sort-array.operator.d.ts.map