import { State } from '../state';
import { DataResult } from '../data-result.interface';
import axios, { CancelToken, CancelTokenSource } from 'axios';
import { DataSource, IdType } from './data-source';
import cloneDeep from 'lodash/cloneDeep';
import { toJS, comparer } from 'mobx';

interface Operations<T, TId> {
    get(state: State, cancelToken?: CancelToken): Promise<DataResult<T>>;
    add?(item: T, index?: number): Promise<T>;
    update?(id: TId, changes: Partial<T>): Promise<void>;
    remove?(id: TId): Promise<T | undefined>;
}

export class AsyncDataSource<T, TId extends IdType = never> implements DataSource<T, TId> {
    idSelector?: (row: T) => TId;

    private operations: Operations<T, TId>;

    private withCache?: boolean;
    private cached?: {
        state: State;
        result: DataResult<T>;
    };

    private cancelation?: CancelTokenSource;

    constructor(operations: Operations<T, TId>, idSelector?: (row: T) => TId, withCache?: boolean) {
        this.operations = operations;
        this.idSelector = idSelector;
        this.withCache = withCache;
    }

    async getData(state: State) {
        if (this.cancelation) {
            this.cancelation.cancel();
        }

        if (this.withCache && this.cached && comparer.structural(this.cached.state, state)) {
            return cloneDeep(this.cached.result);
        }

        this.cancelation = axios.CancelToken.source();
        const result = toJS(await this.operations.get(state, this.cancelation.token));
        this.cancelation = undefined;

        this.cached = { state, result };

        return cloneDeep(result);
    }

    async addData(row: T, index?: number) {
        if (!this.operations.add) {
            throw 'missing add function in the operations';
        }

        const result = await this.operations.add(row, index);

        this.cached = undefined;

        return result;
    }

    async updateData(id: TId, changes: Partial<T>) {
        if (!this.operations.update) {
            throw 'missing update function in the operations';
        }

        await this.operations.update(id, changes);

        this.cached = undefined;
    }

    async removeData(id: TId) {
        if (!this.operations.remove) {
            throw 'missing remove function in the operations';
        }

        const result = await this.operations.remove(id);

        this.cached = undefined;

        return result;
    }
}
